/* 
 * tkMacConsole.c --
 *
 *	This file contains Macintosh-specific procedures for the notifier,
 *	which is the lowest-level part of the Tcl event loop.  This file
 *	works together with ../generic/tclNotify.c.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkMacConsole.c 1.12 96/02/22 10:44:29
 */

#ifndef __CONSOLE__
#include <console.h>
#endif

#include <Events.h>
#include <SIOUX.h> 
#include <GUSI.h> 

#include "tcl.h"
#include "tk.h"
#include "tkMacInt.h"

typedef int (*TclMacConvertEventPtr) _ANSI_ARGS_((EventRecord *eventPtr));

/*
 * Hack to allow Mac Tk to override the TclGetStdChannels function.
 */
 
typedef void (*TclGetStdChannelsProc) _ANSI_ARGS_((Tcl_Channel *stdinPtr,
	Tcl_Channel *stdoutPtr, Tcl_Channel *stderrPtr));
	
EXTERN TclGetStdChannelsProc getStdChannelsProc;

static Tcl_Channel	CreateStdFileChannel _ANSI_ARGS_((char *name,
			    Tcl_File inFile, Tcl_File outFile));
static void		GetStdChannels _ANSI_ARGS_((Tcl_Channel *stdinPtr, 
			    Tcl_Channel *stdoutPtr, Tcl_Channel *stderrPtr));
static int		StdTkIOBlockMode _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, int mode));
static int		StdTkIOBufferSize _ANSI_ARGS_((
    			    ClientData instanceData));
static int		StdTkIOInput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, char *buf, int toRead,
			    int *errorCode));
static int		StdTkIOOutput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File outFile, char *buf, int toWrite,
			    int *errorCode));
static int		StdTkIOClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));

/*
 * The preferred buffer size for StdIO channels.
 */

#define CHANNEL_BUF_SIZE	4096

/*
 * This structure describes the channel type structure for file based IO:
 */

static Tcl_ChannelType consoleChannelType = {
    "file",			/* Type name. */
    StdTkIOBlockMode,		/* Set blocking or
                                 * non-blocking mode.*/
    StdTkIOBufferSize,		/* Buffer size. */
    StdTkIOClose,		/* Close proc. */
    StdTkIOInput,		/* Input proc. */
    StdTkIOOutput,		/* Output proc. */
    NULL			/* Seek proc. */
};

/*
 * Static variables to hold channels for stdin, stdout and stderr.
 */

static Tcl_Channel stdinChannel = NULL;
static Tcl_Channel stdoutChannel = NULL;
static Tcl_Channel stderrChannel = NULL;

/*
 * Static variables to hold files for stdin, stdout and stderr.
 */

static Tcl_File stdinFile = NULL;
static Tcl_File stdoutFile = NULL;
static Tcl_File stderrFile = NULL;
Tcl_Interp *gStdoutInterp = NULL;
static int GUSISpinFunction _ANSI_ARGS_((spin_msg msg, long param));
void 	TclMacSetEventProc _ANSI_ARGS_((TclMacConvertEventPtr procPtr));
int 	TkMacConvertEvent _ANSI_ARGS_((EventRecord *eventPtr));


/*
 *----------------------------------------------------------------------
 *
 * InstallConsole, RemoveConsole, etc. --
 *
 *	The following functions provide the UI for the console package.
 *	Users wishing to replace SIOUX with their own console package 
 *	need only provide the four functions below in a library.
 *
 * Results:
 *	See SIOUX documentation for details.
 *
 * Side effects:
 *	See SIOUX documentation for details.
 *
 *----------------------------------------------------------------------
 */

short 
InstallConsole(short fd)
{
#pragma unused (fd)

	return 0;
}

void 
RemoveConsole(void)
{
}

long 
WriteCharsToConsole(char *buffer, long n)
{
    TkConsolePrint(gStdoutInterp, buffer, n);
    return n;
}

long 
ReadCharsFromConsole(char *buffer, long n)
{
    return 0;
}

extern char *
__ttyname(long fildes)
{
    static char *__devicename = "null device";

    if (fildes >= 0 && fildes <= 2) {
	return (__devicename);
    }
    
    return (0L);
}

short
SIOUXHandleOneEvent(EventRecord *event)
{
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * StdTkIOBufferSize --
 *
 *	Returns the preferred buffer size for unix IO channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdTkIOBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return CHANNEL_BUF_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * StdTkIOOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdTkIOOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    *errorCode = 0;
    errno = 0;
    if (gStdoutInterp != NULL) {
	TkConsolePrint(gStdoutInterp, buf, toWrite);
    }
    
    return toWrite;
}

/*
 *----------------------------------------------------------------------
 *
 * StdTkIOInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

int
StdTkIOInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * StdTkIOBlockMode --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdTkIOBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    return EFAULT;
}

/*
 *----------------------------------------------------------------------
 *
 * StdTkIOClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdTkIOClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* Unused. */
    Tcl_File inFile;		/* Input file to close. */
    Tcl_File outFile;		/* Output file to close. */
{
    int errorCode = 0;

    /*
     * Invalidate the stdio cache if necessary.  Note that we assume that
     * the stdio file and channel pointers will become invalid at the same
     * time.
     */

    if (inFile == stdinFile) {
	stdinFile = NULL;
	stdinChannel = NULL;
    }
    if (outFile == stdoutFile) {
	stdoutFile = NULL;
	stdoutChannel = NULL;
    } else if (outFile == stderrFile) {
	stderrFile = NULL;
	stderrChannel = NULL;
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * GetStdChannels --
 *
 *	Returns handles to the standard channels for stdin, stdout and
 *	stderr.
 *
 * Results:
 *	None; returns values through output parameters.
 *
 * Side effects:
 *	May cause the creation of these standard channels and the underlying
 *	files.
 *
 *----------------------------------------------------------------------
 */

static void
GetStdChannels(stdinPtr, stdoutPtr, stderrPtr)
    Tcl_Channel *stdinPtr,		/* Where to store the channels. */
    	        *stdoutPtr,
    		*stderrPtr;
{
    /*
     * If the channels were not created yet, create them now and
     * store them in the static variables.
     */

    if (stdinPtr != NULL) {
	if (stdinFile == NULL) {
	    stdinFile = Tcl_GetFile((ClientData) 0, TCL_UNIX_FD);
	    stdinChannel = CreateStdFileChannel("stdin", stdinFile, NULL);
        }
        *stdinPtr = (Tcl_Channel) stdinChannel;
    }
    if (stdoutPtr != NULL) {
	if (stdoutFile == NULL) {
	    stdoutFile = Tcl_GetFile((ClientData) 1, TCL_UNIX_FD);
	    stdoutChannel = CreateStdFileChannel("stdout", NULL, stdoutFile);
        }
	*stdoutPtr = (Tcl_Channel) stdoutChannel;
    }
    if (stderrPtr != NULL) {
        if (stderrFile == NULL) {
	    stderrFile = Tcl_GetFile((ClientData) 2, TCL_UNIX_FD);
	    stderrChannel = CreateStdFileChannel("stderr", NULL, stderrFile);
            Tcl_SetChannelOption(NULL, stderrChannel, "-buffering", "none");
        }
        *stderrPtr = (Tcl_Channel) stderrChannel;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CreateStdFileChannel --
 *
 *	Create a channel suitable for use in wrapping a standard
 *	file handle.  Sets up the channel for use by the file driver.
 *
 * Results:
 *	Returns a newly allocated file channel.
 *
 * Side effects:
 *	Registers a new channel with the given name.
 *
 *----------------------------------------------------------------------
 */

static Tcl_Channel
CreateStdFileChannel(name, inFile, outFile)
    char *name;			/* Name of standard channel. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
{
    Tcl_Channel channel;

    channel = Tcl_CreateChannel(&consoleChannelType, name, inFile, outFile,
	    (ClientData) NULL);
    /*
     * Set up the normal channel options for stdio handles.
     */

    if (channel != NULL) {
	Tcl_SetChannelOption(NULL, channel, "-translation", "cr");
	Tcl_SetChannelOption(NULL, channel, "-buffering", "line");
    }
    return channel;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacConsoleInit --
 *
 *	This procedure calls initalization routines to set up a simple
 *	console on a Macintosh.  This is necessary as the Mac doesn't
 *	have a stdout & stderr by default.
 *
 * Results:
 *	Returns TCL_OK if everything went fine.  If it didn't the 
 *	application should probably fail.
 *
 * Side effects:
 *	Inits the appropiate console package.
 *
 *----------------------------------------------------------------------
 */

int
TclMacConsoleInit()
{
    GUSIDefaultSetup();
    GUSISetHook(GUSI_SpinHook, (GUSIHook) GUSISpinFunction);

    TclMacSetEventProc(TkMacConvertEvent);
    getStdChannelsProc = GetStdChannels;

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GUSISpinFunction --
 *
 *	Spins.
 *
 * Results:
 *	Returns .
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
GUSISpinFunction(msg, param)
    spin_msg msg;
    long param;
{
    EventRecord ev;
    
    WaitNextEvent(0, &ev, 0, NULL);
    return 0;
}
void
tkMacInstallMWConsole(Tcl_Interp *interp)
{
    gStdoutInterp = interp;
}
