/* 
 * tkMacNotify.c --
 *
 *	This file provides the platform specific event detection
 *	facilities used by the Tk event routines.  The procedures in
 *	this file comprise the notifier interface and contain unix
 *	specific file and timer handling code.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkMacNotify.c 1.14 96/02/15 18:55:47
 */

#include "tkPort.h"
#include "tkInt.h"
#include "tkMacInt.h"
#include <signal.h>
#include <LowMem.h>

/*
 * A constant the defines the nominal amount of time (in ticks)
 * between calls to WaitNextEvent.
 */

#define WNE_WAIT_TIME 5

/*
 * The information below is used to provide read, write, and
 * exception masks to select during calls to Tcl_DoOneEvent.
 */

static fd_mask checkMasks[3*MASK_SIZE];
				/* This array is used to build up the masks
				 * to be used in the next call to select.
				 * Bits are set in response to calls to
				 * Tcl_WatchFile. */
static fd_mask readyMasks[3*MASK_SIZE];
				/* This array reflects the readable/writable
				 * conditions that were found to exist by the
				 * last call to select. */
static int numFdBits;		/* Number of valid bits in checkMasks
				 * (one more than highest fd for which
				 * Tcl_WatchFile has been called). */

/*
 * The notifier has at most one pending timer.  The next scheduled timer
 * event is recorded in the pendingTimeout structure.  The
 * pendingTimeoutPtr is set to refer to pendingTimeout by Tk_NotifyTimer.
 * Once the timer fires, Tk_DoOneEvent resets the pointer to NULL.
 */

static Tcl_Time *pendingTimeoutPtr;
				/* Points to current timer value. */
static Tcl_Time pendingTimeout;	/* Time when timer should fire. */


/*
 * Prototypes for procedures referenced only in this file:
 */

static void		RefreshFdBits _ANSI_ARGS_((void));

/*
 *----------------------------------------------------------------------
 *
 * Tcl_WatchFile --
 *
 *	Arrange for Tcl_DoOneEvent to include this file in the masks
 *	for the next call to select.  This procedure is invoked by
 *	event sources, which are in turn invoked by Tcl_DoOneEvent
 *	before it invokes select.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	
 *	The notifier will generate a file event when the I/O channel
 *	given by fd next becomes ready in the way indicated by mask.
 *	If fd is already registered then the old mask will be replaced
 *	with the new one.  Once the event is sent, the notifier will
 *	not send any more events about the fd until the next call to
 *	Tcl_NotifyFile. 
 *
 *----------------------------------------------------------------------
 */

void
Tcl_WatchFile(fd, mask)
    int fd;			/* Integer identifier for a stream. */
    int mask;			/* OR'ed combination of TCL_READABLE,
				 * TCL_WRITABLE, and TCL_EXCEPTION:
				 * indicates conditions to wait for
				 * in select. */
{
    int index;
    fd_mask bit;

    index = fd/(NBBY*sizeof(fd_mask));
    bit = 1 << (fd%(NBBY*sizeof(fd_mask)));
    if (mask & TCL_READABLE) {
	checkMasks[index] |= bit;
    }
    if (mask & TCL_WRITABLE) {
	(checkMasks+MASK_SIZE)[index] |= bit;
    }
    if (mask & TCL_EXCEPTION) {
	(checkMasks+2*(MASK_SIZE))[index] |= bit;
    }
    if (numFdBits <= fd) {
	numFdBits = fd+1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FileReady --
 *
 *	Indicates what conditions (readable, writable, etc.) were
 *	present on a file the last time the notifier invoked select.
 *	This procedure is typically invoked by event sources to see
 *	if they should queue events.
 *
 * Results:
 *	The return value is 0 if none of the conditions specified by mask
 *	was true for fd the last time the system checked.  If any of the
 *	conditions were true, then the return value is a mask of those
 *	that were true.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_FileReady(fd, mask)
    int fd;			/* Integer identifier for a stream. */
    int mask;			/* OR'ed combination of TCL_READABLE,
				 * TCL_WRITABLE, and TCL_EXCEPTION:
				 * indicates conditions caller cares about. */
{
    int index, result;
    fd_mask bit;

    index = fd/(NBBY*sizeof(fd_mask));
    bit = 1 << (fd%(NBBY*sizeof(fd_mask)));
    result = 0;
    if ((mask & TCL_READABLE) && (readyMasks[index] & bit)) {
	result |= TCL_READABLE;
    }
    if ((mask & TCL_WRITABLE) && ((readyMasks+MASK_SIZE)[index] & bit)) {
	result |= TCL_WRITABLE;
    }
    if ((mask & TCL_EXCEPTION) && ((readyMasks+(2*MASK_SIZE))[index] & bit)) {
	result |= TCL_EXCEPTION;
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_WaitForEvent --
 *
 *	This procedure does the lowest level wait for events in a
 *	platform-specific manner.  It uses information provided by
 *	previous calls to Tcl_WatchFile, plus the timePtr argument,
 *	to determine what to wait for and how long to wait.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May put the process to sleep for a while, depending on timePtr.
 *	When this procedure returns, an event of interest to the application
 *	has probably, but not necessarily, occurred.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_WaitForEvent(timePtr)
    Tcl_Time *timePtr;		/* Specifies the maximum amount of time
				 * that this procedure should block before
				 * returning.  The time is given as an
				 * interval, not an absolute wakeup time.
				 * NULL means block forever. */
{
    struct timeval timeout, endTime, curTime;
    struct timezone dummyZone;
    int numFound, notDone = true;
    EventRecord macEvent;

    memcpy((VOID *) readyMasks, (VOID *) checkMasks,
	    3*MASK_SIZE*sizeof(fd_mask));
    /*
     * Always call select with a zero timeout.  Calculate the end time.
     */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    gettimeofday(&endTime, &dummyZone);
    endTime.tv_sec += timePtr->sec;
    endTime.tv_usec += timePtr->usec;
    if (endTime.tv_usec > 1000000) {
	endTime.tv_sec++;
    }
    
    do {
	/*
	 * Poll for file events.
	 */
	numFound = select(numFdBits, (SELECT_MASK *) &readyMasks[0],
	    (SELECT_MASK *) &readyMasks[MASK_SIZE],
	    (SELECT_MASK *) &readyMasks[2*MASK_SIZE], &timeout);
	if (numFound > 0) {
	    notDone = false;
	}
	
	/*
	 * Check for time out.
	 */
	if (notDone == true) {
	    gettimeofday(&curTime, &dummyZone);
	    if (curTime.tv_sec >= endTime.tv_sec) {
		if (curTime.tv_sec == endTime.tv_sec) {
		    if (curTime.tv_usec >= endTime.tv_usec) {
			notDone = false;
		    }
		} else {
		    notDone = false;
		}
	    }
	}
	
	/*
	 * Check for window events.
	 */
	if (notDone == true) {
	    if (EventAvail(everyEvent, &macEvent)) {
		notDone = false;
	    }
	}
    } while(notDone == true);
      
    /*
     * Some systems don't clear the masks after an error, so
     * we have to do it here.
     */

    if (numFound == -1) {
	memset((VOID *) readyMasks, 0, 3*MASK_SIZE*sizeof(fd_mask));
    }

    /*
     * Reset the check masks in preparation for the next call to
     * select.
     */

    numFdBits = 0;
    memset((VOID *) checkMasks, 0, 3*MASK_SIZE*sizeof(fd_mask));
}

#ifdef NOT_DEFED
/*
 *----------------------------------------------------------------------
 *
 * Tk_DoOneEvent --
 *
 *	Process a single event of some sort.  If there's no work to
 *	do, wait for an event to occur, then process it.
 *
 * Results:
 *	The return value is 1 if the procedure actually found an event
 *	to process.  If no processing occurred, then 0 is returned.
 *	The caller should invoke Tk_DoOneEvent repeatedly until it
 *	returns 0 to insure that all events have been processed.
 *
 * Side effects:
 *	May delay execution of process while waiting for an event,
 *	unless TK_DONT_WAIT is set in the flags argument.
 *	Tk_ServiceEvent will be called at most once.  If no queued
 *	events were processed, then the procedure checks for new
 *	events.  If TK_DONT_WAIT is set or Tk has called
 *	Tk_NotifyIdle since the last time Tk_ServiceIdle was called,
 *	then Tk_DoOneEvent will only poll for events, otherwise it
 *	blocks until the next event occurs.  All detected events will
 *	be queued.  If no events are detected, and Tk has called
 *	Tk_NotifyIdle, then Tk_ServiceIdle will be called.
 *
 *----------------------------------------------------------------------
 */

int
Tk_DoOneEvent(flags)
    int flags;			/* Miscellaneous flag values:  may be any
				 * combination of TK_DONT_WAIT,
				 * TK_WINDOW_EVENTS, TK_FILE_EVENTS,
				 * TK_TIMER_EVENTS, and TK_IDLE_EVENTS. */
{
    struct timeval curTime, timeout, *timeoutPtr;
    Tk_Event event;
    fd_mask fdmask, bitSelect;
    int index, numFound, waitFlag;
    int foundEvent = 0;
    int dontLoop;		/* Avoids looping if set to 1. */
    long sleepTime;

    /*
     * No event flags is equivalent to TK_ALL_EVENTS.
     */
    
    if ((flags & TK_ALL_EVENTS) == 0) {
	flags |= TK_ALL_EVENTS;
    }

    dontLoop = (flags & TK_DONT_WAIT);

    /*
     * Look for events until we find something to do, or no other events
     * are possible.
     */

    do {

	/*
	 * The first thing we do is to service any asynchronous event
	 * handlers.
	 */
    
	if (Tcl_AsyncReady()) {
	    (void) Tcl_AsyncInvoke((Tcl_Interp *) NULL, 0);
	    return 1;
	}

	/*
	 * Ask Tk to service a queued event.  If Tk does not handle any
	 * events, then we should look for new events.
	 */

	if (Tk_ServiceEvent(flags)) {
	    return 1;		/* Tk serviced an event so we're done. */
	}

	/*
	 * Skip to the idle handlers if that is all we are doing.
	 */

	if ((flags & TK_ALL_EVENTS) == TK_IDLE_EVENTS) {
	    dontLoop = 1;
	    goto doIdle;
	}
    
	/*
	 * Step 1: poll for mouse moved events
	 */

	foundEvent |= TkGeneratePollingEvents();

	/*
	 * Step 2: poll select
	 */

	/*
	 * Load the select mask from the current check mask.  Always call
	 * select with a timeout of zero.  (Always poll.)
	 */

	memcpy((VOID *) ready, (VOID *) check, 3*MASK_SIZE*sizeof(fd_mask));

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	timeoutPtr = &timeout;
	numFound = select(numFdBits, (SELECT_MASK *) &ready[0],
		(SELECT_MASK *) &ready[MASK_SIZE],
		(SELECT_MASK *) &ready[2*MASK_SIZE], timeoutPtr);

	/*
	 * Some systems don't clear the masks after an error, so
	 * we have to do it here.
	 */

	if (numFound == -1) {
	    memset((VOID *) ready, 0, 3*MASK_SIZE*sizeof(fd_mask));
	}

	/*
	 * Check for file events.  We iterate over the ready fd_masks looking
	 * for a mask with any bits set.  By or'ing the read, write, and
	 * exception masks together, we can make a single pass through the file
	 * descriptors.  If any event occured on a file, we need to queue a
	 * file event with the appropriate file descriptor and event mask, and
	 * then clear all of the check bits for that file.
	 */
    
	if (numFound > 0) {
	    foundEvent = 1;
	    event.type = TK_FILE_EVENTS;
	    index = -1;
	    bitSelect = 0;
	    event.file.fd = 0;
	    fdmask = 0;

	    /*
	     * Each time through the loop we shift bitSelect.  When we
	     * have tested each bit in the current mask, bitSelect becomes
	     * zero, so we know to continue with the next mask in the set.
	     */
	
	    while (numFound > 0) {
		if (!bitSelect) {	/* check for wrap-around */
		    bitSelect = 1;
		    index++;
		    fdmask = ready[index] | ready[index+MASK_SIZE]
			| ready[index+2*MASK_SIZE];
		}
		if (fdmask & bitSelect) {
		    numFound--;
		    event.file.mask = 0;
		    if (ready[index] & bitSelect) {
			event.file.mask |= TK_READABLE;
		    }
		    if (ready[index+MASK_SIZE] & bitSelect) {
			event.file.mask |= TK_WRITABLE;
		    }
		    if (ready[index+2*MASK_SIZE] & bitSelect) {
			event.file.mask |= TK_EXCEPTION;
		    }
		    Tk_QueueEvent(&event, TK_QUEUE_TAIL);
		    check[index] &= ~bitSelect;
		    check[index+MASK_SIZE] &= ~bitSelect;
		    check[index+2*MASK_SIZE] &= ~bitSelect;
		}
		bitSelect <<= 1;
		event.file.fd++;
	    }
	}
		
	/* 
	 * Step 3: Window Events & time share 
	 */
	if (foundEvent || ((flags & TK_IDLE_EVENTS) && Tk_IdlePending())) {
	    sleepTime = 0;
	    waitFlag = TK_DONT_WAIT;
	} else {
	    waitFlag = 0;
	    if (pendingTimeoutPtr == NULL || pendingTimeoutPtr->sec > 1) {
		sleepTime = WNE_WAIT_TIME;
	    } else {
		sleepTime = pendingTimeoutPtr->usec / 16666;
		if (sleepTime > WNE_WAIT_TIME) {
		    sleepTime = WNE_WAIT_TIME;
		} else if (sleepTime == 0) {
		    waitFlag = TK_DONT_WAIT;
		}
	    }
	}
	foundEvent |= TkHandleMacEvents(waitFlag, sleepTime);
    
	/*
	 * Step 4: Check to see if the timer has expired.
	 */

	if ((pendingTimeoutPtr != NULL) && (flags & TK_TIMER_EVENTS)) {
	    (void) gettimeofday(&curTime, (struct timezone *) NULL);
	    if ((pendingTimeoutPtr->sec < curTime.tv_sec)
		    || ((pendingTimeoutPtr->sec == curTime.tv_sec)
			    &&  (pendingTimeoutPtr->usec < curTime.tv_usec))) {
		pendingTimeoutPtr = NULL;

		/*
		 * Generate timer event.
		 */

		foundEvent = 1;
		event.type = TK_TIMER_EVENTS;
		event.timer.time.sec = curTime.tv_sec;
		event.timer.time.usec = curTime.tv_usec;
		Tk_QueueEvent(&event, TK_QUEUE_TAIL);
	    }
	}

	/*
	 * If no other events were generated, then we should service any
	 * pending idle handlers.
	 */

	doIdle:
	if (!foundEvent && (flags & TK_IDLE_EVENTS)) {
	    foundEvent = Tk_ServiceIdle();
	}
    } while (!foundEvent && !dontLoop);

    /*
     * We found nothing at all to do this time around.
     */
    return foundEvent;
}
#endif

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Sleep --
 *
 *	Delay execution for the specified number of milliseconds.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Time passes.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_Sleep(ms)
    int ms;			/* Number of milliseconds to sleep. */
{
    long endTime;

    if (ms <= 0) {
	return;
    }
	
    /* 
     * One tick == 16.66 milliseconds 
     */
    Delay( (ms / 16.66), &endTime);
}
